from webthing import (SingleThing, Property, Thing, Value, WebThingServer)
from RPLCD.i2c import CharLCD, BaseCharLCD
from pi_display_webthing.display import Display
import tornado.ioloop
import logging


class DisplayWebThing(Thing):
    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing

    def __init__(self, name: str, description: str, lcd: BaseCharLCD):
        Thing.__init__(
            self,
            'urn:dev:ops:lcddisplay-1',
            'Display ' + name + ' Controller',
            ['Display'],
            description
        )

        self.display = Display(lcd, self.__update_text)

        self.display_text = Value("")
        self.add_property(
            Property(self,
                     'displayed_text',
                     self.display_text,
                     metadata={
                         'title': 'Displayed text',
                         'type': 'string',
                         'description': 'Displayed text',
                         'readOnly': True,
                     }))

        self.upper_layer_text = Value("", self.display.panel(Display.LAYER_UPPER).update_text)
        self.add_property(
            Property(self,
                     'upper_layer_text',
                     self.upper_layer_text,
                     metadata={
                         'title': 'Upper layer text',
                         'type': 'string',
                         'description': 'The text of the upper layer',
                         'readOnly': False,
                     }))

        self.upper_layer_text_ttl = Value(-1, self.display.panel(Display.LAYER_UPPER).update_ttl)
        self.add_property(
            Property(self,
                     'upper_layer_text_ttl',
                     self.upper_layer_text_ttl,
                     metadata={
                         'title': 'Upper layer text (time-to-live)',
                         'type': 'integer',
                         'description': 'The time-to-live of the upper layer. Value -1 deactivates ttl',
                         'readOnly': False,
                     }))

        self.middle_layer_text = Value("", self.display.panel(Display.LAYER_MIDDLE).update_text)
        self.add_property(
            Property(self,
                     'middle_layer_text',
                     self.middle_layer_text,
                     metadata={
                         'title': 'Middle layer text',
                         'type': 'string',
                         'description': 'The text of the middle layer',
                         'readOnly': False,
                     }))

        self.middle_layer_text_ttl = Value(-1, self.display.panel(Display.LAYER_MIDDLE).update_ttl)
        self.add_property(
            Property(self,
                     'middle_layer_text_ttl',
                     self.middle_layer_text_ttl,
                     metadata={
                         'title': 'Middle layer text (time-to-live)',
                         'type': 'integer',
                         'description': 'The time-to-live of the middle layer. Value -1 deactivates ttl',
                         'readOnly': False,
                     }))

        self.lower_layer_text = Value("", self.display.panel(Display.LAYER_LOWER).update_text)
        self.add_property(
            Property(self,
                     'lower_layer_text',
                     self.lower_layer_text,
                     metadata={
                         'title': 'Lower layer text',
                         'type': 'string',
                         'description': 'The text of the lower layer',
                         'readOnly': False,
                     }))

        self.lower_layer_text_ttl = Value(-1, self.display.panel(Display.LAYER_LOWER).update_ttl)
        self.add_property(
            Property(self,
                     'lower_layer_text_ttl',
                     self.lower_layer_text_ttl,
                     metadata={
                         'title': 'Lower layer text (time-to-live)',
                         'type': 'integer',
                         'description': 'The time-to-live of the lower layer. Value -1 deactivates ttl',
                         'readOnly': False,
                     }))

        self.ioloop = tornado.ioloop.IOLoop.current()


    def __update_text(self):
        self.ioloop.add_callback(self.__update_text_props)

    def __update_text_props(self):
        self.display_text.notify_of_external_update(self.display.text)
        self.upper_layer_text.notify_of_external_update(self.display.panel(Display.LAYER_UPPER).text)
        self.upper_layer_text_ttl.notify_of_external_update(self.display.panel(Display.LAYER_UPPER).ttl)
        self.middle_layer_text.notify_of_external_update(self.display.panel(Display.LAYER_MIDDLE).text)
        self.middle_layer_text_ttl.notify_of_external_update(self.display.panel(Display.LAYER_MIDDLE).ttl)
        self.lower_layer_text.notify_of_external_update(self.display.panel(Display.LAYER_LOWER).text)
        self.lower_layer_text_ttl.notify_of_external_update(self.display.panel(Display.LAYER_LOWER).ttl)


def createI2C(i2c_expander: str, i2c_address: int) -> BaseCharLCD:
    logging.info("bind driver to address " + hex(i2c_address) + " using port expander " + i2c_expander)
    return CharLCD(i2c_expander, i2c_address)


def run_server(port: int, name:str, i2c_expander: str, i2c_address: int, description: str):
    lcd = createI2C(i2c_expander, i2c_address)
    display_webthing = DisplayWebThing(name, description, lcd)
    server = WebThingServer(SingleThing(display_webthing), port=port, disable_host_validation=True)
    try:
        logging.info('starting the server')
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping the server')
        server.stop()
        logging.info('done')



from RPLCD.i2c import BaseCharLCD
import threading
import time
import uuid


class Panel:

    def __init__(self, display, changed_listener):
        self.display = display
        self.changed_listener = changed_listener
        self.__text = ""
        self.__ttl = -1
        self.ttl_watchdog_id = None

    @property
    def ttl(self) -> int:
        return self.__ttl

    @property
    def text(self) -> str:
        return self.__text

    def update_text(self, text: str):
        self.__text = text
        self.display.on_panel_updated()
        self.changed_listener()

    def update_ttl(self, ttl: int):
        self.__ttl = ttl
        wdid = str(uuid.uuid4())
        self.ttl_watchdog_id = wdid
        threading.Thread(target = self.run_expire_watchdog, args = (ttl, wdid)).start()

    def run_expire_watchdog(self, ttl: int, wdid: str):
        remaining_sec = ttl
        while (remaining_sec > 0):
            time.sleep(1)
            remaining_sec = remaining_sec - 1
            if self.ttl_watchdog_id == wdid:
                if remaining_sec > 0:
                    self.__ttl = remaining_sec
                    self.changed_listener()
                else:
                    self.clear()

    def clear(self):
        self.__text = ""
        self.__ttl = -1
        self.display.on_panel_updated()
        self.changed_listener()

    def is_empty(self):
        return len(self.__text) == 0


class Display:

    LAYER_UPPER = 0
    LAYER_MIDDLE = 1
    LAYER_LOWER = 2

    def __init__(self, lcd: BaseCharLCD, changed_listener):
        self.__lcd = lcd
        self.__text = ""
        self.changed_listener = changed_listener
        self.panels = [Panel(self, changed_listener), Panel(self, changed_listener), Panel(self, changed_listener)]

    def panel(self, layer: int) -> Panel:
        return self.panels[layer]

    @property
    def text(self) -> str:
        return self.__text

    def __update_text(self, text: str):
        self.__text = text
        self.__lcd.clear()
        self.__lcd.write_string(self.__text)
        self.changed_listener()

    def on_panel_updated(self):
        text = ""
        for layer in range(0, len(self.panels)):
            panel = self.panels[layer]
            if not panel.is_empty():
                text = panel.text
                break
        self.__update_text(text)


from os import system, remove
from os import listdir
from abc import ABC
import pathlib
import logging
import subprocess
import argparse



class App(ABC):

    def __init__(self, packagename: str, entrypoint: str, description: str):
        self.unit = Unit(packagename)
        self.packagename = packagename
        self.entrypoint = entrypoint
        self.description = description

    def do_add_argument(self, parser):
        pass

    def do_process_command(self, command:str, port: int, verbose: bool, args) -> bool:
        return False

    def do_additional_listen_example_params(self):
        return ""

    def print_usage_info(self, port: str, msg: str=None):
        if msg is not None:
            print(msg + "\n")

        if port is None:
            port = "9496"

        print("for command options usage")
        print(" sudo " + self.entrypoint + " --help")
        print("example commands")
        print(" sudo " + self.entrypoint + " --command register --port " + port + " " + self.do_additional_listen_example_params())
        print(" sudo " + self.entrypoint + " --command listen --port " + port + " " + self.do_additional_listen_example_params())
        if len(self.unit.list_installed()) > 0:
            print("example commands for registered services")
            for service_info in self.unit.list_installed():
                port = service_info[1]

                print(" sudo " + self.entrypoint + " --command deregister --port " + port)
                print(" sudo " + self.entrypoint + " --command log --port " + port)


    def handle_command(self):
        parser = argparse.ArgumentParser(description=self.description)
        parser.add_argument('--command', metavar='command', required=False, type=str, help='the command. Supported commands are: listen (run the webthing service), register (register and starts the webthing service as a systemd unit, deregister (deregisters the systemd unit), log (prints the log)')
        parser.add_argument('--port', metavar='port', required=False, type=int, help='the port of the webthing serivce')
        parser.add_argument('--verbose', metavar='verbose', required=False, type=bool, default=False, help='activates verbose output')
        self.do_add_argument(parser)
        args = parser.parse_args()

        if args.verbose:
            log_level=logging.DEBUG
        else:
            log_level=logging.INFO
        logging.basicConfig(format='%(asctime)s %(name)-20s: %(levelname)-8s %(message)s', level=log_level, datefmt='%Y-%m-%d %H:%M:%S')
        print("using log level " + str(log_level))

        if args.command is None:
            self.print_usage_info(str(args.port))
        elif args.command == 'deregister':
            if args.port is None:
                self.print_usage_info(None, "--port is mandatory for deregister command")
            else:
                self.unit.deregister(int(args.port))
        elif args.command == 'log':
            if args.port is None:
                self.print_usage_info(str(args.port), "--port is mandatory for log command")
            else:
                self.unit.printlog(int(args.port))
        else:
            if args.port is not None:
                if self.do_process_command(args.command, args.port, args.verbose, args):
                    return
            self.print_usage_info(str(args.port))



class Unit:

    def __init__(self, packagename: str):
        self.packagename = packagename

    def register(self, port: int, unit: str):
        service = self.servicename(port)
        unit_file_fullname = str(pathlib.Path("/", "etc", "systemd", "system", service))
        with open(unit_file_fullname, "w") as file:
            file.write(unit)
        system("sudo systemctl daemon-reload")
        system("sudo systemctl enable " + service)
        system("sudo systemctl restart " + service)
        system("sudo systemctl status " + service)

    def deregister(self, port: int):
        service = self.servicename(port)
        unit_file_fullname = str(pathlib.Path("/", "etc", "systemd", "system", service))
        system("sudo systemctl stop " + service)
        system("sudo systemctl disable " + service)
        system("sudo systemctl daemon-reload")
        try:
            remove(unit_file_fullname)
        except Exception as e:
            pass

    def printlog(self, port:int):
        system("sudo journalctl -f -u " + self.servicename(port))

    def servicename(self, port: int):
        return self.packagename + "_" + str(port) + ".service"

    def list_installed(self):
        services = []
        try:
            for file in listdir(pathlib.Path("/", "etc", "systemd", "system")):
                if file.startswith(self.packagename) and file.endswith('.service'):
                    idx = file.rindex('_')
                    port = file[idx+1:file.index('.service')]
                    services.append((file, port, self.is_active(file)))
        except Exception as e:
            pass
        return services

    def is_active(self, servicename: str):
        cmd = '/bin/systemctl status %s' % servicename
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE,encoding='utf8')
        stdout_list = proc.communicate()[0].split('\n')
        for line in stdout_list:
            if 'Active:' in line:
                if '(running)' in line:
                    return True
        return False



from smbus2 import SMBus
from typing import List


def scan() -> List[str]:
    devices = []
    try:
        bus = SMBus(1) # 1 indicates /dev/i2c-1
        for device in range(128):
            try:
                bus.read_byte(device)
                devices.append(hex(device))
            except:
                pass
    except FileNotFoundError as e:
        print("WARNING: I2C seems not to be activated")
    return devices




from string import Template
from pi_display_webthing.app import App
from pi_display_webthing.display_webthing import run_server
from pi_display_webthing.detect_devices import scan


PACKAGENAME = 'pi_display_webthing'
ENTRY_POINT = "display"
DESCRIPTION = "A web connected LCD display module"


UNIT_TEMPLATE = Template('''
[Unit]
Description=$packagename
After=syslog.target network.target

[Service]
Type=simple
ExecStart=$entrypoint --command listen --verbose $verbose --port $port --name $name --i2c_address $i2c_address --i2c_expander $i2c_expander
SyslogIdentifier=$packagename
StandardOutput=syslog
StandardError=syslog
Restart=always
RestartSec=30

[Install]
WantedBy=multi-user.target
''')




class DhtApp(App):

    def do_add_argument(self, parser):
        parser.add_argument('--i2c_expander', metavar='i2c_expander', required=False, type=str, help='the name of the port I²C port expander. Supported: PCF8574, MCP23008, MCP23017')
        parser.add_argument('--i2c_address', metavar='i2c_address', required=False, type=str, help='the I²C address of the LCD Module as hex string')
        parser.add_argument('--name', metavar='name', required=False, type=str, default="", help='the name of the display')

    def do_additional_listen_example_params(self):
        devices = scan()
        if len(devices) > 0:
            device = devices[0]
        else:
            device = " 0x29"
        return "--name nas --i2c_expander PCF8574 --i2c_address " + device

    def do_process_command(self, command:str, port: int, verbose: bool, args) -> bool:
        if command in {'listen', 'register'} and args.i2c_address is None:
            devices = scan()
            if len(devices) == 0:
                print("no devices detected")
                return False
            elif len(devices) > 1:
                print("more than 1 devices found: " + ", ".join(devices) + "\n use --i2c_address parameter to select device")
                return False
            else:
                args.i2c_address = devices[0]
                print("device " + args.i2c_address + " detected. Using it")

        if command == 'listen' and (args.i2c_expander is not None) and (args.i2c_address is not None):
            print("running " + self.packagename + " on  " + str(port))
            run_server(port, args.name, args.i2c_expander, self.to_hex(args.i2c_address), self.description)
            return True

        elif args.command == 'register' and (args.i2c_expander is not None) and (args.i2c_address is not None):
            print("register " + self.packagename  + " on " + str(port) + " and starting it")
            unit = UNIT_TEMPLATE.substitute(packagename=self.packagename, entrypoint=self.entrypoint, port=port, verbose=verbose, name=args.name, i2c_expander=args.i2c_expander, i2c_address=args.i2c_address)
            self.unit.register(port, unit)
            return True
        else:
            return False

    def to_hex(self, hexString: str) -> int:
        if hexString.startswith("0x"):
            return int(hexString, 16)
        else:
            return int(hexString)


def main():
    DhtApp(PACKAGENAME, ENTRY_POINT, DESCRIPTION).handle_command()


if __name__ == '__main__':
    main()
